; -------------------------------------------------------------------------
; KeyRand.Asm by Robert Rothenburg Walking-Owl <rrothenb@ic.sunysb.edu>
; -------------------------------------------------------------------------
; This routine latches onto the keyboard interrupt serivce routine (Int 09)
; and adds keystroke timings to a random number pool. This is presumably a
; real and unpredictable source of randomness for applications which need
; it (such as crypto).
;
; The advantage of latching onto the keyboard interrupt is that a program
; can collect random bits (or bytes) in the background from the user's key-
; strokes.
;
; No guarantees or warrantees of any sort are made on this software, implied
; or otherwise.  Nor will the author be responsible for any damages that may
; result from using this software: in other words, use at your own risk.
;
; This software is Copyright (c) 1994, although you may freely distribute,
; use and or modify it so long as this copyright notice is not removed from 
; the source.  If you distribute a modified version, please do so with
; commented code noting the modifications and date, etc.
;
;        4 Dec. 1994  - initial publication (R.Walking-Owl)
;
; The procedures in this routine were written to be incorporated into Pascal 
; programs, although it should be no trouble to use them in C (either define
; the routines in a header file with "pascal" calls or put underscores prior
; to the procedure names, whatever suits your taste...
;
; A brief description of the functions follows:
;
;   InstallRandom returns -1 if successful, 0 if not; UnInstallRandom does 
;       the same: their purpose should be obvious.  Remember to have your
;       Ctrl-Break handler uninstall the interrupt!
;
;   GetRandPoolAddr (only is __publicpool is defined) returns the address of
;       the random data pool.  This may have to be rewritten if huge or flat
;       memory-models are used.  (Adjust the RandPoolSize symbol to suit 
;       your needs.)
;
;   Noise returns 8-bits of noise from the random number pool.
;
;   NumRandBits returns the number of random bits in the pool.
;
; No arguments are required for any of these routines, so they should be
; fairly portable across compilers. The source has been tested with both
; MASM 5.0 and TASM (version?) and assembles with no problems.
;
; -------------------------------------------------------------------------


;__for286            equ         1  ; target for 286+ machines

randpoolsize        equ        32   ; size of random number pool we want
__samplebits        equ         1   ; sample bits, not bytes
__publicpool        equ         1   ; define if randpool is to be public

dos                 equ         21h  ; ms-dos interrupt vector                   
keybd0              equ         60h  ; keyboard port

                    ifdef       __for286
                      .286
                    endif

code                segment     byte public
                    assume      cs: code
                    assume      ds: nothing, es: nothing, ss: nothing

; -------------------------------------------------------------------------
; installs the handler and initializes stuff
; -------------------------------------------------------------------------
               public  installrandom
installrandom  proc    far
                pushf
                cli
                push    ds
                call    checkforinst
                or      ax,     ax
                jnz     @handlerisinstalled
                
                mov     ax,             3509h ; get int 09 address
                int     dos
                
                cmp     bx, offset randomhandler ; double check!
                jne     @handlernotinstalled
                mov     ax, es
                mov     dx, cs
                cmp     ax, dx
                jne     @handlernotinstalled
@handlerisinstalled: 
                xor     ax,     ax
                jmp     short   @exitinstall
@handlernotinstalled:                
                mov     word ptr cs: [ oldint09vector+2 ], es
                mov     word ptr cs: [ oldint09vector ], bx
                mov     ax,             2509h ; set new int 09 address
                mov     dx,             cs
                mov     ds,             dx
                mov     dx,             offset randomhandler
                int     dos
                call  pctimer0
                xchg  ax, word ptr cs: [lasttick]
                call    checkforinst ; verify install
@exitinstall:   pop     ds
                popf
                ret
installrandom  endp

; -------------------------------------------------------------------------
; uninstalls the handler               
; -------------------------------------------------------------------------
; note: rememer to have ctrl-break or critical-error routines call this                
;       procedure!

                public  uninstallrandom
uninstallrandom proc   far
                pushf
                cli
                push    ds
                
                call    checkforinst   ; are we installed?
                or      ax,  ax
                jz      @cantuninstall

                mov     ax,             2509h
                mov     dx,             word ptr cs: [ oldint09vector ]
                mov     ds,             word ptr cs: [ oldint09vector+2 ]
                int     dos
                call    checkforinst    ; verify it!
                not     ax
@cantuninstall: pop     ds
                popf
                ret
uninstallrandom endp

; -------------------------------------------------------------------------
; check to see if installed Int 09 handler is at vector (ax is nz. if ok)               
; -------------------------------------------------------------------------

; Note: this routine checks the addresses at the interrupt vector, and not
;       for a marker at that address which is required should this code be
;       incorporated into a TSR.

; Because this routine is automagically called by the install and uninstall
; procedures, there is no need for it to be public.  You can change that if
; you'd like...

checkforinst    proc   near
                pushf   ; we want to preserve cli just in case
                cli
                mov     dx,             -1
                xor     ax,             ax
                mov     es,             ax
                mov     ax,             offset randomhandler
                cmp     word ptr es: [ 0024h ], ax
                jne     short @latch
                mov     ax,             cs
                cmp     word ptr es: [ 0026h ], ax
                je      @nolatch
@latch:         not     dx
@nolatch:       mov     ax,             dx
                popf
                ret
checkforinst    endp

oldint09vector  label   dword     ; where the original vector is
                dd      (?)

; -------------------------------------------------------------------------
; the random sampling handler
; -------------------------------------------------------------------------
                  
randomhandler   proc    far
                pushf
                cli
                push    ax
                push    bx
               
                  push  si
                  call  pctimer0
                  mov   bx,     ax
                  xchg  ax,     word ptr cs: [lasttick]
                  sub   bx,     ax
                 
              ifndef    __samplebits    
                  mov   si,     (randpoolsize-1)
                  xor   bl,     byte ptr cs: [randompool][si]
@swirlrandompoolbytes:         
                  mov   al,     byte ptr cs: [randompool][si-1]
                  mov   byte ptr cs: [randompool][si], al
                  dec   si
                  or    si,     si
                  jnz   @swirlrandompoolbytes  
                  mov   byte ptr cs: [randompool], bl
                  add   word ptr cs: [bitsinpool], 8
               else
                  xor   si,     si
                  push  cx
                  mov   cx,     randpoolsize
                  rcr   bx,     1
@swirlrandompoolbits:
                  rcr   byte ptr cs: [randompool][si], 1
                  inc   si
                  loop  @swirlrandompoolbits 
                  inc   word ptr cs: [bitsinpool]
                  pop   cx
               endif

                 cmp    word ptr cs: [bitsinpool], (randpoolsize*8)
                 jna    noadjust
                 mov    word ptr cs: [bitsinpool], (randpoolsize*8)
noadjust:

                  pop   si
                
                in      al,     keybd0

                pop     bx
                pop     ax

                popf
                jmp dword ptr cs: [ oldint09vector ]
randomhandler   endp
                

; -------------------------------------------------------------------------
; some routines to poll the system timer for some random data
; -------------------------------------------------------------------------
; These routines are loosely based on similarly named procedures in the
; PGP 2.6.2 sources.  A more detailed explanation of what they do can be
; found there.
                    
; A potential modification would be to change the system timer interrupt
; so that it is called every few ms. rather than 18.2 times per second:
; one would also have to make appropriate changes to the system timer
; routines so that the clock doesn't advance several hours per minute...
                    
                    if  randpoolsize

timer0              equ         40h                        ;
timercntl           equ         43h                        ;
                    
; -------------------------------------------------------------------------
; routine to sample the pc timer port
; -------------------------------------------------------------------------
; based on the pctimer0() function in noise.c from the pgp 2.6.2 sources

pctimer0            proc        near
                    push        bx
                    ifndef      __for286
                      push        cx
                    endif
                    in          al,         timer0
                    mov         al,         0c2h
                    out         timercntl,  al
                    in          al,         timer0
                    and         al,         80h
                    mov         bx,         ax
                    ifndef      __for286     
                      mov       cl,         8 
                      shl       bx,         cl    
                    else
                      shl       bx,         8     
                    endif
                    xor         al,         al
                    out         timercntl,  al
                    in          al,         timer0
                    and         ax,         0ffh
                    shr         ax,         1
                    or          bx,         ax
                    in          al,         timer0
                    and         ax,         0ffh
                    ifndef      __for286
                      mov       cl,         7             
                      shl       ax,         cl           
                    else
                      shl       ax,         7              
                    endif
                      or        ax,         bx
                    ifndef      __for286
                      pop       cx
                    endif
                    pop         bx
                    ret         
pctimer0            endp        

bitsinpool          label       word
                    dw          0

lasttick            label       word  ; save value of last call to pctimer0
                    dw          (?)

                    endif

; -------------------------------------------------------------------------
; the random-number pool...
; -------------------------------------------------------------------------
; Note: one could maintain a random-seed file and load RandPoolSize bytes
;       into the randompool (prior to initializing the handler, pref.)

randompool            label byte
                      db    randpoolsize dup (?)

; -------------------------------------------------------------------------
; returns 8 bits of noise
; -------------------------------------------------------------------------
; a routine could be written to extract n bits of noise from the random
; pool as well...
                    public noise
noise               proc   far
                    xor    al, al
                    cmp    word ptr cs: [bitsinpool], 8
                    jb     nobitsleft
                    mov    ah, byte ptr cs:[randompool]
                    xor    si, si
shiftpooldown:      mov    al, byte ptr cs: [randompool][si+1]
                    mov    byte ptr cs: [randompool][si], al
                    inc    si
                    cmp    si, randpoolsize
                    jb     shiftpooldown
                    sub    word ptr cs: [bitsinpool], 8
                    mov    al, ah
nobitsleft:                    
                    ret
noise               endp

; -------------------------------------------------------------------------
; returns the number of random bits in the pool
; -------------------------------------------------------------------------
                    public numrandbits
numrandbits         proc   far
                    mov    ax, word ptr cs: [bitsinpool]
                    ret
numrandbits         endp

; -------------------------------------------------------------------------
; returns the address of the randompool
; -------------------------------------------------------------------------
; note: it's probably a good idea to wipe the random pool when you are                    
;       done with it. (the uninstall procedure doesn't do that!)

                  ifdef __publicpool  
                    public  getrandpooladdr
getrandpooladdr     proc    far
                    mov     dx, cs
                    mov     ax, offset randompool
                    ret
getrandpooladdr     endp
                  endif

; A particularly psychotic idea might be to have a TSR collect random data
; from keystroke timings into a huge file while the user does other normal
; things like typing DOS commands or using a word processor.

code                ends        
                  end         
